from argparse import ArgumentParser
from asyncio import CancelledError, Runner, Semaphore, Task, TaskGroup
from dataclasses import dataclass
from pathlib import Path
from shutil import copytree, rmtree
from time import time_ns
from typing import Any

from yaml import safe_dump

from debatrix.core.common import DebateResult, Verdict
from debatrix.platform import Platform, Session


@dataclass
class ScriptArgs:
    motion: str = ""
    record_file: str = ""
    repeat: int = 1

    api_key: str | None = None
    model: str | None = None
    base_url: str | None = None
    proxy: str | None = None


async def work(session: Session, /, *, args: ScriptArgs) -> None:
    openai_config: dict[str, Any] = session.config_data["model"]["chat_config"]["openai_config"]

    if args.api_key is not None:
        openai_config["api_key"] = args.api_key
    if args.model is not None:
        openai_config["model"] = args.model
    if args.base_url is not None:
        openai_config["base_url"] = args.base_url
    if args.proxy is not None:
        openai_config["proxy"] = args.proxy

    await session.update_config()

    motion_code: str = session.motions[0][0]
    await session.select_debate(motion_code)
    assert session.cur_info is not None

    title: str = f"{motion_code} --- {session.cur_info.motion}"
    print(f"Start: {title}\n")

    try:
        result: DebateResult | None = await session.start_debate()
        print(f"End: {title}\n")

        if result is None:
            print("Debate cancelled.\n")
        else:
            final_verdict: Verdict | None = result.final_verdict
            assert final_verdict is not None

            for debater_verdict in final_verdict.debaters_verdict:
                print(f"{debater_verdict.debater_name}: {debater_verdict.score}")

            print(f"Winner: {final_verdict.winner_verdict.winner}\n")

        print(f"Verdict saved at {await session.save_record()}\n")
    except Exception as e:
        print(f"Failed: {e}\n")
    finally:
        await session.reset_debate()


async def main(platform: Platform, /, *, args: ScriptArgs) -> None:
    async with TaskGroup() as tg:
        task: Task[None] = tg.create_task(platform.serve())

        try:
            semaphore = Semaphore(12)

            async def wait_and_work() -> None:
                async with semaphore:
                    await work(await platform.assign(), args=args)

            async with TaskGroup() as inner_tg:
                for _ in range(args.repeat):
                    inner_tg.create_task(wait_and_work())
        finally:
            task.cancel()

            try:
                await task
            except CancelledError:
                pass


def extract(fp: str) -> list[dict[str, str]]:
    records: list[dict[str, str]] = []
    buffer: list[str] = []

    def clear_buffer() -> None:
        if len(buffer) > 0:
            records.append({"debater_name": buffer[1][6:8], "content": "\n".join(buffer[2:])[8:]})
            buffer.clear()

    with open(fp, "r", encoding="utf8") as f:
        for line in f:
            if line.startswith("2024-"):
                clear_buffer()

            buffer.append(line.strip())

    clear_buffer()
    return records


if __name__ == "__main__":
    arg_parser = ArgumentParser(description="Debatrix for CCAC")

    arg_parser.add_argument("motion", help="debate motion")
    arg_parser.add_argument("record_file", help="debate record file")

    arg_parser.add_argument(
        "-r", "--repeat", type=int, default=1, help="judge repetition (can judge simultaneously)"
    )

    arg_parser.add_argument("-k", "--api_key", help="API key for any OpenAI-style API")
    arg_parser.add_argument("-m", "--model", help="model name (default: gpt-3.5-turbo)")

    arg_parser.add_argument(
        "-u", "--base-url", help="(optional) base URL for other OpenAI-style APIs"
    )

    arg_parser.add_argument("-x", "--proxy", help="(optional) proxy for API connection")

    args: ScriptArgs = arg_parser.parse_args(namespace=ScriptArgs())
    debate: list[dict[str, str]] = extract(args.record_file)

    resource_path: Path = Path("./output") / f"{Path(args.record_file).stem}_{time_ns()}"
    resource_path.mkdir(parents=True, exist_ok=True)
    print(f"Using resource path: {resource_path.as_posix()}")

    config_path: Path = resource_path / "config"
    motion_path: Path = resource_path / "motion"
    speech_path: Path = resource_path / "speech"

    for target_path in (config_path, motion_path, speech_path):
        if target_path.exists():
            if target_path.is_symlink() or not target_path.is_dir():
                target_path.unlink()
            else:
                rmtree(target_path, ignore_errors=True)

    print("Loading config & debate ...")
    copytree(Path("./config"), config_path)
    motion_path.mkdir(parents=True, exist_ok=True)
    speech_path.mkdir(parents=True, exist_ok=True)

    debate_name: str = "local"
    with (motion_path / "motion_list.txt").open("w", encoding="utf8") as f:
        print(f"{debate_name}\t{args.motion}", file=f)

    with (motion_path / f"{debate_name}.yml").open("w", encoding="utf8") as f:
        safe_dump(
            {
                "motion": args.motion,
                "pro_side": [{"name": "正方"}],
                "con_side": [{"name": "反方"}],
                "info_slide": "No information.",
                "speech_order": [speech["debater_name"] for speech in debate],
            },
            f,
            allow_unicode=True,
            sort_keys=False,
        )

    with (speech_path / f"{debate_name}.yml").open("w", encoding="utf8") as f:
        safe_dump(debate, f, allow_unicode=True, sort_keys=False)

    Runner(debug=True).run(main(Platform(resource_path, fast_api_log_info=False), args=args))

    print("Cleaning up ...")
    for target_path in (config_path, motion_path, speech_path):
        rmtree(target_path)
